/*
 * Copyright (C)  NHR@FAU, University Erlangen-Nuremberg.
 * All rights reserved. This file is part of nusif-solver.
 * Use of this source code is governed by a MIT style
 * license that can be found in the LICENSE file.
 */
#include <stdio.h>
#include <stdlib.h>

#include "comm.h"

#if defined(_MPI)
// subroutines local to this module
static int sizeOfRank(int rank, int size, int N)
{
    return N / size + ((N % size > rank) ? 1 : 0);
}

static int sum(int* sizes, int position)
{
    int sum = 0;

    for (int i = 0; i < position; i++) {
        sum += sizes[i];
    }

    return sum;
}

static void gatherArray(
    Comm* c, int cnt, int* rcvCounts, int* displs, double* src, double* dst)
{
    double* sendbuffer = src + (c->imaxLocal + 2);

    if (c->rank == 0) {
        sendbuffer = src;
    }

    MPI_Gatherv(sendbuffer,
        cnt,
        MPI_DOUBLE,
        dst,
        rcvCounts,
        displs,
        MPI_DOUBLE,
        0,
        MPI_COMM_WORLD);
}
#endif // defined _MPI

// exported subroutines
void commReduction(double* v, int op)
{
#if defined(_MPI)
    if (op == MAX) {
        MPI_Allreduce(MPI_IN_PLACE, v, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
    } else if (op == SUM) {
        MPI_Allreduce(MPI_IN_PLACE, v, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
    }
#endif
}

int commIsBoundary(Comm* c, int direction)
{
#if defined(_MPI)
    switch (direction) {
    case LEFT:
        return 1;
        break;
    case RIGHT:
        return 1;
        break;
    case BOTTOM:
        return c->rank == 0;
        break;
    case TOP:
        return c->rank == (c->size - 1);
        break;
    }
#endif

    return 1;
}

void commExchange(Comm* c, double* grid)
{
#if defined(_MPI)
    MPI_Request requests[4] = { MPI_REQUEST_NULL,
        MPI_REQUEST_NULL,
        MPI_REQUEST_NULL,
        MPI_REQUEST_NULL };

    /* exchange ghost cells with top neighbor */
    if (c->rank + 1 < c->size) {
        int top     = c->rank + 1;
        double* src = grid + (c->jmaxLocal) * (c->imaxLocal + 2) + 1;
        double* dst = grid + (c->jmaxLocal + 1) * (c->imaxLocal + 2) + 1;

        MPI_Isend(src, c->imaxLocal, MPI_DOUBLE, top, 1, MPI_COMM_WORLD, &requests[0]);
        MPI_Irecv(dst, c->imaxLocal, MPI_DOUBLE, top, 2, MPI_COMM_WORLD, &requests[1]);
    }

    /* exchange ghost cells with bottom neighbor */
    if (c->rank > 0) {
        int bottom  = c->rank - 1;
        double* src = grid + (c->imaxLocal + 2) + 1;
        double* dst = grid + 1;

        MPI_Isend(src, c->imaxLocal, MPI_DOUBLE, bottom, 2, MPI_COMM_WORLD, &requests[2]);
        MPI_Irecv(dst, c->imaxLocal, MPI_DOUBLE, bottom, 1, MPI_COMM_WORLD, &requests[3]);
    }

    MPI_Waitall(4, requests, MPI_STATUSES_IGNORE);
#endif
}

void commShift(Comm* c, double* f, double* g)
{
#if defined(_MPI)
    MPI_Request requests[2] = { MPI_REQUEST_NULL, MPI_REQUEST_NULL };

    /* shift G */
    /* receive ghost cells from bottom neighbor */
    if (c->rank > 0) {
        int bottom = c->rank - 1;
        MPI_Irecv(g + 1,
            c->imaxLocal,
            MPI_DOUBLE,
            bottom,
            0,
            MPI_COMM_WORLD,
            &requests[0]);
    }

    if (c->rank + 1 < c->size) {
        int top     = c->rank + 1;
        double* buf = g + (c->jmaxLocal) * (c->imaxLocal + 2) + 1;
        /* send ghost cells to top neighbor */
        MPI_Isend(buf, c->imaxLocal, MPI_DOUBLE, top, 0, MPI_COMM_WORLD, &requests[1]);
    }

    MPI_Waitall(2, requests, MPI_STATUSES_IGNORE);
#endif
}

// TODO Merge with seq
void commCollectResult(Comm* c,
    double* ug,
    double* vg,
    double* pg,
    double* u,
    double* v,
    double* p,
    int jmax,
    int imax)
{
    int *rcvCounts, *displs;
    int cnt = c->jmaxLocal * (imax + 2);

    if (c->rank == 0) {
        rcvCounts = (int*)malloc(c->size * sizeof(int));
        displs    = (int*)malloc(c->size * sizeof(int));
    }

    if (c->rank == 0 && c->size == 1) {
        cnt = (c->jmaxLocal + 2) * (imax + 2);
    } else if (c->rank == 0 || c->rank == (c->size - 1)) {
        cnt = (c->jmaxLocal + 1) * (imax + 2);
    }

    MPI_Gather(&cnt, 1, MPI_INTEGER, rcvCounts, 1, MPI_INTEGER, 0, MPI_COMM_WORLD);

    if (c->rank == 0) {
        displs[0]  = 0;
        int cursor = rcvCounts[0];

        for (int i = 1; i < c->size; i++) {
            displs[i] = cursor;
            cursor += rcvCounts[i];
        }
    }

    gatherArray(c, cnt, rcvCounts, displs, p, pg);
    gatherArray(c, cnt, rcvCounts, displs, u, ug);
    gatherArray(c, cnt, rcvCounts, displs, v, vg);
}

void commPrintConfig(Comm* c)
{
#if defined(_MPI)
    fflush(stdout);
    MPI_Barrier(MPI_COMM_WORLD);
    if (commIsMaster(c)) {
        printf("Communication setup:\n");
    }

    for (int i = 0; i < c->size; i++) {
        if (i == c->rank) {
            printf("\tRank %d of %d\n", c->rank, c->size);
            printf("\tNeighbours (bottom, top, left, right): %d %d, %d, %d\n",
                c->neighbours[BOTTOM],
                c->neighbours[TOP],
                c->neighbours[LEFT],
                c->neighbours[RIGHT]);
            printf("\tCoordinates (j,i) %d %d\n", c->coords[JDIM], c->coords[IDIM]);
            printf("\tLocal domain size (j,i) %dx%d\n", c->jmaxLocal, c->imaxLocal);
            fflush(stdout);
        }
    }
    MPI_Barrier(MPI_COMM_WORLD);
#endif
}

void commInit(Comm* c, int argc, char** argv)
{
#if defined(_MPI)
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &(c->rank));
    MPI_Comm_size(MPI_COMM_WORLD, &(c->size));
#else
    c->rank = 0;
    c->size = 1;
#endif
}

void commPartition(Comm* c, int jmax, int imax)
{
#if defined(_MPI)
    c->imaxLocal = imax;
    c->jmaxLocal = sizeOfRank(c->rank, c->size, jmax);
#else
    c->imaxLocal = imax;
    c->jmaxLocal = jmax;
#endif
}

void commFinalize(Comm* c)
{
#if defined(_MPI)
    for (int i = 0; i < NDIRS; i++) {
        MPI_Type_free(&c->sbufferTypes[i]);
        MPI_Type_free(&c->rbufferTypes[i]);
    }

    MPI_Finalize();
#endif
}
